; Frequency Meter, Version 1.4, 14 March 2000

; COPYRIGHT (C) 2000, T. M. Napier.

; Written for a 20.000 MHz clock, 200 nS/instruction


; This program switches between counting and displaying.  While
; counting it tracks the content of the PIC's timer register and
; increments a further register on each overflow.  The count
; period is automatically selected to be from 10 mS to 10 sec
; as appropriate for the input pulse rate.
; During counting the timer register is sampled every 50 uS.
; This time is set by the loop instruction count, 250 instructions.

; At the end of a count period the result is converted from binary
; to decimal and displayed for three seconds on the multiplexed
; LED display.  During this time the ports generate the segment
; drive output while cycling the digit drive shift register chip.
; The display update is based on the 2 mS dwell time per digit.
; 250 cycles around all six digits is exactly three seconds.

; The input count is copied to three registers REG1-REG3 which are
; left shifted into six digit registers during binary to decimal
; conversion.  Any overflows detected during counting or conversion
; trigger an upward (shorter count) range change.  If the most
; significant display digit is a zero the range is changed down.

; Input/Output
;	Port A  Four bit prescaler input

;	Port B  Segment drive, bit 0 = A segment, bit 7 = DP

;	Port C	Bit 0	not used (my chip is faulty)
;		Bit 1	Count on/off		out
;		Bit 2	Prescaler clear		out
;		Bit 3	Shift register data	out
;		Bit 4	Shift register clock	out
;		Bit 5	Display mode switch	in
;		Bit 6	not used		in
;		Bit 7	Test flag		out

;PIC register allocation

TEMP	EQU	8	;Temporary buffer
DIGT	EQU	9	;Digit counter
LOOP	EQU	10	;Loop counter
LAST	EQU	11	;Last timer count
HIGH	EQU	12	;High byte of input count

REG1	EQU	13	;Shift register for binary to decimal
REG2	EQU	14
REG3	EQU	15	;Low byte of shift register

DIG1	EQU	16	;High digit of display
DIG2	EQU	17
DIG3	EQU	18
DIG4	EQU	19
DIG5	EQU	20
DIG6	EQU	21	;Low digit of display

; The delay time registers time both the
; count period and the display period.

TIME0	EQU	22	;Counts microseconds to 50
TIME1	EQU	23	;Counts 200 loops to get 10 mS
TIME2	EQU	24	;Counts 1 to 1000 outer loops
TIME3	EQU	25	;Upper byte of outer loop count

RANGE	EQU	26	;Period selector, 1 - 4
FLAGS	EQU	27	;Flag bits
DP	EQU	28	;Shift register for decimal point info

;PORT C bit allocations

RUN	EQU	1	;high = counting
RESET	EQU	2	;Low = reset prescaler
DATIN	EQU	3	;Data pin of shift register
CLOCK	EQU	4	;Clock pin of shift register
SWIT	EQU	5	;Mode switch input

;Bits in FLAGS register

OVER1	EQU	0	;Pulse count overflow
OVER2	EQU	1	;Conversion overflow
SW1	EQU	2	;Set if switch down
SW2	EQU	3	;Set if switch up
SWF	EQU	4	;Shows if count done

NORM	EQU	47	;Normal loop count for 50 uS

;Dumb stuff the assembler should know already.

INDEX	EQU	0
RTCC	EQU	1
PC	EQU	2
STATUS	EQU	3
IREG	EQU	4
PORTA	EQU	5
PORTB	EQU	6
PORTC	EQU	7

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

;Start of program code

	MOVLW	0FFH
	TRIS	PORTA		;Set Port A to all input
	MOVWF	PORTB		;All display segments off
	MOVLW	0
	TRIS	PORTB		;Set Port B to all output
	MOVWF	PORTC		;Prescaler reset, count off
	MOVLW	61H
	TRIS	PORTC		;Port C, two input pins
	MOVLW	0F8H
	OPTION			;Set direct timer input
	MOVLW	1
	MOVWF	RANGE		;Set 50 MHz range
	CLRF	FLAGS		;No overflows or switch input

	MOVLW	8
	MOVWF	DIGT
ST1	CALL	DSHFT
	DECFSZ	DIGT,1
	GOTO	ST1		;Clear display shift register

	CALL	DIGSET		;Set count and pointer
ST2	CLRF	INDEX		;Clear a digit
	INCF	IREG,1		;Next digit
	DECFSZ	DIGT,1
	GOTO	ST2

	GOTO	SHOW		;Display = 000000

;Lookup table converts digits to segments
;Segment is on if bit is low

CONV	ANDLW	15		;safety precaution
	ADDWF	PC,1
	RETLW	0C0H		;Segments for 0
	RETLW	0F9H		;Segments for 1
	RETLW	0A4H		;Segments for 2
	RETLW	0B0H		;Segments for 3
	RETLW	099H		;Segments for 4
	RETLW	092H		;Segments for 5
	RETLW	082H		;Segments for 6
	RETLW	0F8H		;Segments for 7
	RETLW	080H		;Segments for 8
	RETLW	098H		;Segments for 9
	RETLW	0BFH		;Overflow, show -
	RETLW	0BFH		
	RETLW	0BFH		
	RETLW	0BFH		
	RETLW	0BFH		
	RETLW	0FFH		;Blank

;Lookup table for decimal point position. Also sets MHz and kHz LEDs.

POINT	DECF	RANGE,0		;W = (Range - 1) = 0-3
	ADDWF	PC,1
	RETLW	22H
	RETLW	21H
	RETLW	14H
	RETLW	12H

;Set up loop count and index to address six digits

DIGSET	MOVLW	6
	MOVWF	DIGT		;Digit counter
	MOVLW	DIG1
	MOVWF	IREG		;Initial data address
	RETLW	0

;Display all six digits once.  Monitors mode switch.

DISP	CALL	POINT		;Convert range to point info
	MOVWF	DP		;Save result
	BSF	PORTC,DATIN	;Set for first digit only
	CALL	DIGSET
DSP1	MOVF	INDEX,0		;Get digit
	CALL	CONV		;Convert it to segments
	MOVWF	PORTB		;Output segment pattern
	BTFSC	DP,0		;Test bit in DP register
	BCF	PORTB,7		;Turn on decimal point or LED
	RRF	DP,1		;Shift ready for next digit 
	CALL	DSHFT		;Clock the digit select register
	BCF	PORTC,DATIN	;Clear for all digits after first

;2 mS hold for each digit

	MOVLW	10	
	MOVWF	TIME2
WT2	MOVLW	200		;200 uS delay
	MOVWF	TIME1
WT1	NOP
	NOP
	DECFSZ	TIME1,1
	GOTO	WT1		;1 uS loop

	DECFSZ	TIME2,1
	GOTO	WT2		;200 uS loop

;Read position of mode switch during digits 1 and 2

	MOVF	DIGT,0
	XORLW	6		;Test if digit 1
	BTFSS	STATUS,2
	GOTO	DSP2

	BCF	FLAGS,SW1
	BTFSC	PORTC,SWIT
	BSF	FLAGS,SW1	;Record switch position
	GOTO	DSP3

DSP2	MOVF	DIGT,0
	XORLW	5		;Test if digit 2
	BTFSS	STATUS,2
	GOTO	DSP3

	BCF	FLAGS,SW2
	BTFSC	PORTC,SWIT
	BSF	FLAGS,SW2	;Record switch position

DSP3	INCF	IREG,1		;Next digit
	DECFSZ	DIGT,1
	GOTO	DSP1

DSHFT	BSF	PORTC,CLOCK
	NOP
	BCF	PORTC,CLOCK	;Turn off all digits
	RETLW	0

;Convert count in REG1-3 to decimal in DIG1-6
;REG1 = High byte, REG2 = Middle byte, REG3 = Low byte = prescaler
;Bottom four bits of REG3 are zeros.

;Algorithm  For each of 20 input bits, double the decimal number,
; adding 1 if the input bit is a 1.  Maintain digits in 0-9 range
; by subtracting 10 and carrying into higher digits as needed.

B2DEC	CALL	DIGSET
BD1	CLRF	INDEX		;Clear a digit register
	INCF	IREG,1
	DECFSZ	DIGT,1
	GOTO	BD1

;Do bit-by-bit conversion
	MOVLW	20
	MOVWF	LOOP		;Bit counter
BD2	MOVLW	6
	MOVWF	DIGT
	MOVLW	DIG6
	MOVWF	IREG		;Address lowest digit
	BCF	STATUS,0	;Avoid garbage in register
	RLF	REG3,1		;Shift input number left
	RLF	REG2,1
	RLF	REG1,1		;Puts input msb in carry

;Double each digit, add carry from previous step then decimal correct

BD3	RLF	INDEX,1		;Double digit and add carry
	MOVLW	10
	SUBWF	INDEX,0		;Carry set if digit > 9
	BTFSS	STATUS,0
	GOTO	BD4

;If digit > 9 then subtract 10 and carry into next digit

	MOVLW	10		;Since W was clobbered by SUBWF
	SUBWF	INDEX,1		;Subtract for real this time
BD4	DECF	IREG,1
	DECFSZ	DIGT,1		;Test if done
	GOTO	BD3

	BTFSC	STATUS,0	;Carry left by last subtraction
	BSF	FLAGS,OVER2	;Flag any overflow from the last digit
	DECFSZ	LOOP,1
	GOTO	BD2

	RETLW	0

;Timing routine
; TIME0 counts microseconds to make a 50 uS inner loop.
; TIME1 counts 200 inner loops to make a 10 mS loop.
; TIME2 and TIME3 count 1, 10, 100 or 1000 of 10 mS loops.
; Each successive loop shortens the first 50 uS period
; to make time for decrementing its loop counter.

AGAIN	MOVF	RANGE,0
	MOVWF	TEMP		;Copy RANGE
	CLRF	TIME3
	MOVLW	1		;Set 10 mS as default
	DECF	TEMP,1
	BTFSC	STATUS,2
	GOTO	DOIT

	MOVLW	10		;Set 100 mS
	DECF	TEMP,1
	BTFSC	STATUS,2
	GOTO	DOIT
	
	MOVLW	100		;Set 1 sec
	DECF	TEMP,1
	BTFSC	STATUS,2
	GOTO	DOIT
	
	MOVLW	3
	MOVWF	TIME3
	MOVLW	232		;Set 10 sec
DOIT	MOVWF	TIME2		;Save low byte of outer loop count
	MOVLW	0FCH
	ANDWF	FLAGS,1		;Reset overflow flags
	BCF	PORTC,RESET	;Send reset pulse to prescaler
	BSF	PORTC,RESET
	CLRF	LAST		;Tracks last timer content to check overflow
	CLRF	HIGH		;Counts timer overflows
	CLRF	RTCC		;Clear timer register
;Set initial loop counts
	MOVLW	200
	MOVWF	TIME1		;Set 10 mS loop count
	MOVLW	NORM-2
	MOVWF	TIME0		;Set shorter 50 uS loop count
	BSF	PORTC,RUN	;Start count, Instruction count> 0

;Enter inner loop with loop count in TIME0

DELAY	NOP				;			 1
	NOP				;			 2
	DECFSZ	TIME0,1			;			 3
	GOTO	DELAY		;1 uS loop			 4

; Sample counter register every 50 uS.
; If this count < last count then do carry into next register.

	MOVF	RTCC,0		;Get current count		 5
	MOVWF	TEMP		;Save it			 6
	MOVF	LAST,0		;Get last count			 7
	SUBWF	TEMP,0		;Set carry			 8
	BTFSC	STATUS,0		;			 9
	GOTO	DEL2		;Waste some time		10

	INCF	HIGH,1		;Increment if overflow		11
	BTFSC	STATUS,2		;			12
	BSF	FLAGS,OVER1	;HIGH register overflowed	13
DEL1	MOVF	TEMP,0		;Get current count		14
	MOVWF	LAST		;Save it as last count		15
	MOVLW	NORM		;Length of regular loop		16
	MOVWF	TIME0		;Set normal loop count		17
	DECFSZ	TIME1,1			;			18
	GOTO	DELAY		;Continue for 200 loops		19
					;		GOTO =  20
; Decrement the outer loop count.

	MOVLW	200			;			20
	MOVWF	TIME1		;Reset 10 mS counter		21
	DECF	TIME0,1		;Set 50 uS counter to NORM-1	22
	DECFSZ	TIME2,1			;			23
	GOTO	DELAY		;Non-zero so carry on looping	24
					;		GOTO =  25

	DECF	TIME0,1		;Saves doing it later		25
	MOVF	TIME3,1		;It might already be zero	26
	BTFSC	STATUS,2				;			27
	GOTO	DONE			;			28
					;		GOTO =  29

	DECF	TIME0,1		;Set 50 uS counter to NORM-3	29
	DECF	TIME3,1			;			30
	NOP				;			31
	NOP				;			32
	NOP				;			33
	GOTO	DELAY		;Not done yet			34
					;		GOTO =  35

DEL2	GOTO	DEL1		;Add some delay			12
					;		GOTO =  13

DONE	BCF	PORTC,RUN	;Stop count			30

;Check timer byte for one last overflow

	MOVF	LAST,0		;Get last count
	SUBWF	RTCC,1		;Set carry
	BTFSC	STATUS,0
	GOTO	DON1

	INCF	HIGH,1		;Increment if overflow
	BTFSC	STATUS,2
	BSF	FLAGS,OVER1

;Set up for decimal conversion

DON1	MOVF	HIGH,0
	MOVWF	REG1		;High byte of count
	MOVF	RTCC,0		;Get timer byte
	MOVWF	REG2		;Middle byte of count
	SWAPF	PORTA,0		;Read prescaler into high nibble
	MOVWF	REG3
	CALL	B2DEC		;Do decimal conversion

;Autoranging  If overflow, decrement RANGE but not past 1.
;If most significant digit is zero, Increment RANGE but not past 4.
; If the range is unsuitable a new count is done before anything
; is displayed.

	MOVF	FLAGS,0
	ANDLW	3		;Check for overflow
	BTFSS	STATUS,2
	GOTO	RANDN		;Shorten sampling period

	MOVF	DIG1,0		;Check for underflow (MSD = 0)
	BTFSS	STATUS,2
	GOTO	SHOW		;Since range OK

	MOVF	RANGE,0
	XORLW	4
	BTFSC	STATUS,2
	GOTO	SHOW		;Since range can't change

	INCF	RANGE,1		;Increment RANGE unless it's 4
	GOTO	AGAIN		;Do count with new range

RANDN	MOVF	RANGE,0
	XORLW	1
	BTFSC	STATUS,2
	GOTO	SHOW		;Since range can't change

	DECF	RANGE,1		;Decrement RANGE unless it's 1	
	GOTO	AGAIN		;Do count with new range

;Display the result.
; If switch up then do 250 loops (3 sec) then exit.
; If switch center then display indefinitely.
; If switch down do one count and display indefinitely.

SHOW	MOVLW	250
	MOVWF	LOOP		;3 second timer
SHW1	CALL	DISP		;Display six digits once
	BTFSS	FLAGS,SW2	;In auto mode?
	GOTO	SHW2

	DECFSZ	LOOP,1		;Do count-down
	GOTO	SHW1		;Continue display

	GOTO	AGAIN		;Do another count

SHW2	BTFSS	FLAGS,SW1	;In hold mode?
	GOTO	SHW3

	BTFSC	FLAGS,SWF	;Test if count already taken
	GOTO	SHOW		;Took count so just display

	BSF	FLAGS,SWF
	MOVLW	1
	MOVWF	RANGE		;Set range for quick reaction
	GOTO	AGAIN		;Take one count

;Center position so clear flag and continue to display count

SHW3	BCF	FLAGS,SWF
	GOTO	SHOW

	END
